from datetime import datetime, timedelta, timezone
from io import StringIO
from textwrap import dedent
from unittest.mock import MagicMock, call, patch, seal

from botocore.exceptions import ClientError
from django.core.management import call_command
from django.test import override_settings

from arkindex.documents.models import CorpusExport, CorpusExportState, Element
from arkindex.images.models import Image, ImageServer
from arkindex.ponos.models import Artifact, Farm, Task
from arkindex.process.models import (
    DataFile,
    Process,
    ProcessMode,
    Worker,
    WorkerRun,
    WorkerType,
    WorkerVersion,
)
from arkindex.project.tests import FixtureTestCase
from arkindex.training.models import Dataset, Model, ModelVersion


@override_settings(AWS_EXPORT_BUCKET="export", PONOS_S3_ARTIFACTS_BUCKET="ponos-artifacts", PONOS_S3_LOGS_BUCKET="ponos-logs", AWS_TRAINING_BUCKET="training")
@patch("django_rq.get_connection")
@patch("arkindex.project.aws.s3")
class TestCleanupCommand(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker = Worker.objects.create(
            name="A worker",
            slug="test_worker",
            type=WorkerType.objects.first(),
            repository_url="http://nerv.co.jp/nerv/eva/",
        )

    def cleanup(self, args=[]):
        """
        Run the cleanup command, capturing its output
        """
        output = StringIO()
        call_command("cleanup", args + ["--no-color"], stdout=output, stderr=output)
        return output.getvalue().strip()

    def _make_artifact(self):
        """
        Create an artifact on an expired process
        """
        process = Process.objects.create(
            farm=Farm.objects.first(),
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.superuser,
        )
        task = process.tasks.create(
            run=0,
            depth=0,
            slug="task",
            expiry=datetime(1970, 1, 1, tzinfo=timezone.utc),
            ttl=0,
        )
        return task.artifacts.create(path="artichoke", size=99999)

    def test_cleanup(self, s3_mock, rq_mock):
        young_export = self.corpus.exports.create(user=self.superuser)
        # Use a fake creation time to make old exports
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(days=42)
            self.corpus.exports.create(user=self.user, state=CorpusExportState.Failed)
            done_export = self.corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)

        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 1 old corpus exports from S3…
                Removing export {done_export.id} from S3…
                Removing 2 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        self.assertEqual(CorpusExport.objects.count(), 1)
        self.assertEqual(CorpusExport.objects.get(), young_export)
        self.assertEqual(s3_mock.Object.call_count, 1)
        self.assertEqual(s3_mock.Object.call_args, call("export", str(done_export.id)))
        self.assertEqual(s3_mock.Object.return_value.delete.call_count, 1)

    def test_nothing(self, s3_mock, rq_mock):
        self.assertFalse(CorpusExport.objects.exists())
        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

    def test_s3_not_found(self, s3_mock, rq_mock):
        s3_mock.Object.return_value.delete.side_effect = ClientError({"Error": {"Code": "404"}}, "delete_object")
        seal(s3_mock)

        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(days=42)
            done_export = self.corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)

        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 1 old corpus exports from S3…
                Removing export {done_export.id} from S3…
                Export {done_export.id} not found on S3, skipping
                Removing 1 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        self.assertFalse(CorpusExport.objects.exists())
        self.assertEqual(s3_mock.Object.call_count, 1)
        self.assertEqual(s3_mock.Object.call_args, call("export", str(done_export.id)))
        # s3.Object.delete should only be called once, not retried
        self.assertEqual(s3_mock.Object.return_value.delete.call_count, 1)

    def test_s3_error(self, s3_mock, rq_mock):
        error = ClientError({"Error": {"Code": "500"}}, "delete_object")
        # Fail twice, then delete successfully
        s3_mock.Object.return_value.delete.side_effect = [error, error, None]
        seal(s3_mock)

        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(days=42)
            done_export = self.corpus.exports.create(user=self.superuser, state=CorpusExportState.Done)

        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 1 old corpus exports from S3…
                Removing export {done_export.id} from S3…
                Removing 1 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        self.assertFalse(CorpusExport.objects.exists())
        self.assertEqual(s3_mock.Object.call_count, 1)
        self.assertEqual(s3_mock.Object.call_args, call("export", str(done_export.id)))
        self.assertEqual(s3_mock.Object.return_value.delete.call_count, 3)

    def test_cleanup_trashed_datafiles(self, s3_mock, rq_mock):
        DataFile.objects.bulk_create([
            DataFile(
                name=f"test{i}.txt",
                size=42,
                content_type="text/plain",
                corpus=self.corpus,
                trashed=True if i % 2 else False
            ) for i in range(0, 2)
        ])

        trashed_df = DataFile.objects.filter(trashed=True).get()
        self.assertEqual(DataFile.objects.count(), 2)
        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 1 DataFiles marked as trashed from S3 and the database…
                Deleting DataFile {trashed_df.id}…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )
        self.assertEqual(DataFile.objects.count(), 1)
        self.assertEqual(DataFile.objects.filter(trashed=True).count(), 0)
        self.assertEqual(s3_mock.Object().delete.call_count, 1)

    def test_cleanup_trashed_datafiles_ignore_s3_errors(self, s3_mock, rq_mock):
        """
        Test the cleanup command tries multiple times to delete from S3, but ignores errors if it fails
        """
        DataFile.objects.bulk_create([
            DataFile(
                name=f"test{i}.txt",
                size=42,
                content_type="text/plain",
                corpus=self.corpus,
                trashed=True if i % 2 else False
            ) for i in range(0, 2)
        ])

        s3_mock.Object().delete.side_effect = [
            ClientError({"Error": {"Code": "500"}}, "delete_object"),
            ClientError({"Error": {"Code": "500"}}, "delete_object"),
            ValueError("Something went wrong"),
        ]
        seal(s3_mock)

        trashed_df = DataFile.objects.filter(trashed=True).get()
        self.assertEqual(DataFile.objects.count(), 2)
        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 1 DataFiles marked as trashed from S3 and the database…
                Deleting DataFile {trashed_df.id}…
                Something went wrong
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )
        self.assertEqual(DataFile.objects.count(), 2)
        self.assertEqual(DataFile.objects.filter(trashed=True).count(), 1)
        self.assertEqual(s3_mock.Object().delete.call_count, 3)

    @patch("arkindex.documents.management.commands.cleanup.s3")
    def test_cleanup_artifacts(self, cleanup_s3_mock, s3_mock, rq_mock):
        process = Process.objects.create(
            farm=Farm.objects.first(),
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.superuser,
        )
        task = process.tasks.create(run=0, depth=0, slug="task", ttl=0)

        good_s3_artifact = MagicMock()
        good_s3_artifact.key = f"{task.id}/path/to/thing.txt"
        orphan_s3_artifact = MagicMock()
        orphan_s3_artifact.key = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/sad/artifact.zip"
        unsupported_s3_artifact = MagicMock()
        unsupported_s3_artifact.key = "cant_touch_this.txt.vbs"
        broken_s3_artifact = MagicMock()
        broken_s3_artifact.key = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/nope.zip"
        broken_s3_artifact.delete.side_effect = ClientError({"Error": {"Code": "500"}}, "delete_object")

        cleanup_s3_mock.Bucket.return_value.objects.all.side_effect = [
            # Bucket for Ponos artifacts
            [good_s3_artifact, orphan_s3_artifact, unsupported_s3_artifact, broken_s3_artifact],
            # Bucket for corpus exports
            [],
            # Bucket for IIIF images
            [],
            # Bucket for Ponos logs
            [],
            # Bucket for model version archives
            []
        ]

        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Removing artifact aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/sad/artifact.zip…
                Unsupported artifact cant_touch_this.txt.vbs
                Removing artifact aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa/nope.zip…
                An error occurred (500) when calling the delete_object operation: Unknown
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )
        self.assertEqual(cleanup_s3_mock.Bucket.call_count, 5)
        self.assertListEqual(cleanup_s3_mock.Bucket.call_args_list, [
            call("ponos-artifacts"),
            call("export"),
            call("iiif"),
            call("ponos-logs"),
            call("training")
        ])
        self.assertEqual(cleanup_s3_mock.Bucket().objects.all.call_count, 5)
        self.assertEqual(good_s3_artifact.delete.call_count, 0)
        self.assertEqual(orphan_s3_artifact.delete.call_count, 1)
        self.assertEqual(unsupported_s3_artifact.delete.call_count, 0)
        self.assertEqual(broken_s3_artifact.delete.call_count, 1)

    def test_cleanup_expired_processes(self, s3_mock, rq_mock):
        farm = Farm.objects.first()

        expired_process = farm.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.superuser,
        )
        expired_task = expired_process.tasks.create(
            run=0,
            depth=0,
            slug="task",
            expiry=datetime(1970, 1, 1, tzinfo=timezone.utc),
            ttl=0,
        )
        expired_artifact = expired_task.artifacts.create(path="nope.txt", size=256)

        expired_process_2 = farm.processes.create(
            mode=ProcessMode.Dataset,
            corpus=self.corpus,
            creator=self.superuser,
        )
        expired_task_with_dataset_link = expired_process_2.tasks.create(
            run=0,
            depth=0,
            slug="task",
            expiry=datetime(1970, 1, 1, tzinfo=timezone.utc),
            ttl=0,
        )
        expired_task_with_dataset_link.dataset.set([
            Dataset.objects.create(name="SEELE", description="Neon Genesis Evangelion", corpus=self.corpus, creator=self.user)
        ])
        expired_artifact_with_dataset_link = expired_task_with_dataset_link.artifacts.create(path="nope.txt", size=256)

        non_expired_process = farm.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.superuser,
        )
        non_expired_task = non_expired_process.tasks.create(
            run=0,
            depth=0,
            slug="task",
            # Expiration date in the future: not yet expired
            expiry=datetime.now(timezone.utc) + timedelta(days=99),
            ttl=0,
        )
        non_expired_artifact = non_expired_task.artifacts.create(path="artsy-fact", size=1337)

        # A process with no tasks at all never expires
        empty_process = farm.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.superuser,
        )

        s3_mock.Object().key = "s3_key"
        s3_mock.Object().delete.return_value = None
        seal(s3_mock)

        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 1 artifacts of expired processes from S3…
                Removing artifact s3_key
                Removing logs for 1 tasks of expired processes from S3…
                Removing task log s3_key
                Removing 1 artifacts of expired processes…
                Removing 1 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        with self.assertRaises(Task.DoesNotExist):
            expired_task.refresh_from_db()
        with self.assertRaises(Artifact.DoesNotExist):
            expired_artifact.refresh_from_db()

        # Those still exist, refreshing works
        expired_process.refresh_from_db()
        expired_process_2.refresh_from_db()
        non_expired_process.refresh_from_db()
        expired_task_with_dataset_link.refresh_from_db()
        non_expired_task.refresh_from_db()
        expired_artifact_with_dataset_link.refresh_from_db()
        non_expired_artifact.refresh_from_db()
        empty_process.refresh_from_db()

        self.assertEqual(s3_mock.Object().delete.call_count, 2)

    def test_cleanup_expired_processes_null(self, s3_mock, rq_mock):
        # This revision on the `main` branch does not have any WorkerVersions.
        # Improper handling of NULL values in the queries looking for expired processes and
        # excluding revisions that should not be deleted could lead to
        # this revision causing no expired processes to ever be found.

        # This WorkerVersion with no docker_image could break the query.
        self.worker.versions.create(configuration={}, version=1)

        # These artifacts should be cleaned up
        lonely_artifact = self._make_artifact()
        lonely_artifact_2 = self._make_artifact()

        s3_mock.Object().key = "s3_key"
        s3_mock.Object().delete.return_value = None
        seal(s3_mock)

        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 2 artifacts of expired processes from S3…
                Removing artifact s3_key
                Removing artifact s3_key
                Removing logs for 2 tasks of expired processes from S3…
                Removing task log s3_key
                Removing task log s3_key
                Removing 2 artifacts of expired processes…
                Removing 2 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        with self.assertRaises(Artifact.DoesNotExist):
            lonely_artifact.refresh_from_db()
        with self.assertRaises(Task.DoesNotExist):
            lonely_artifact.task.refresh_from_db()
        with self.assertRaises(Artifact.DoesNotExist):
            lonely_artifact_2.refresh_from_db()
        with self.assertRaises(Task.DoesNotExist):
            lonely_artifact_2.task.refresh_from_db()

        # Those still exist, refreshing works
        lonely_artifact.task.process.refresh_from_db()
        lonely_artifact_2.task.process.refresh_from_db()

        self.assertEqual(s3_mock.Object().delete.call_count, 4)

    @patch("arkindex.documents.management.commands.cleanup.s3")
    def test_cleanup_local_images(self, cleanup_s3_mock, s3_mock, rq_mock):
        ImageServer.objects.local.images.create(path="path%2Fto%2Fimage.jpg")
        img_object = MagicMock()
        img_object.key = "path/to/image.jpg"
        broken_object = MagicMock()
        broken_object.key = "thisisfine.jpg"
        broken_object.delete.side_effect = ClientError({"Error": {"Code": "500"}}, "delete_object")
        orphan_object = MagicMock()
        orphan_object.key = "jeanpaul2.jp2"

        cleanup_s3_mock.Bucket.return_value.objects.all.side_effect = [
            # Bucket for Ponos artifacts
            [],
            # Bucket for corpus exports
            [],
            # Bucket for IIIF images
            [broken_object, img_object, orphan_object],
            # Bucket for Ponos logs
            [],
            # Bucket for model version archives
            []
        ]

        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Removing image thisisfine.jpg…
                An error occurred (500) when calling the delete_object operation: Unknown
                Removing image jeanpaul2.jp2…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )
        self.assertEqual(cleanup_s3_mock.Bucket.call_count, 5)
        self.assertListEqual(cleanup_s3_mock.Bucket.call_args_list, [
            call("ponos-artifacts"),
            call("export"),
            call("iiif"),
            call("ponos-logs"),
            call("training")
        ])
        self.assertEqual(cleanup_s3_mock.Bucket().objects.all.call_count, 5)
        self.assertEqual(img_object.delete.call_count, 0)
        self.assertEqual(broken_object.delete.call_count, 1)
        self.assertEqual(orphan_object.delete.call_count, 1)

    @patch("arkindex.documents.management.commands.cleanup.s3")
    def test_cleanup_orphan_images(self, cleanup_s3_mock, s3_mock, rq_mock):
        element = Element.objects.get(name="Volume 2, page 1v")
        image_with_element = Image.objects.get(id=element.image.id)
        img_server = ImageServer.objects.get(url="http://server")
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(weeks=3)
            image_no_element_old_1 = Image.objects.create(path="path/path/img", width=12, height=12, server=img_server)
            image_no_element_old_2 = Image.objects.create(path="path/pathpathpath/img", width=12, height=12, server=img_server)
        image_no_element_new = Image.objects.create(path="path/pathpath/img", width=12, height=12, server=img_server)

        with self.assertNumQueries(21):
            self.assertEqual(
                self.cleanup(),
                dedent(
                    """
                    Removing archived workers…
                    Successfully cleaned up 0 archived workers.
                    Removing orphaned Ponos artifacts…
                    Successfully cleaned up orphaned Ponos artifacts.
                    Removing 0 artifacts of expired processes from S3…
                    Removing logs for 0 tasks of expired processes from S3…
                    Removing 0 artifacts of expired processes…
                    Removing 0 tasks of expired processes…
                    Successfully cleaned up expired processes.
                    Removing 0 old corpus exports from S3…
                    Removing 0 old corpus exports…
                    Successfully cleaned up old corpus exports.
                    Removing orphaned corpus exports…
                    Successfully cleaned up orphaned corpus exports.
                    Deleting 0 DataFiles marked as trashed from S3 and the database…
                    Successfully cleaned up DataFiles marked as trashed.
                    Removing orphan images…
                    Removing 2 orphan images…
                    Successfully cleaned up orphan images.
                    Removing orphaned local images…
                    Successfully cleaned up orphaned local images.
                    Removing orphaned Ponos logs…
                    Successfully cleaned up orphaned Ponos logs.
                    Removing archived models…
                    Successfully cleaned up 0 archived models.
                    Removing orphaned model versions archives…
                    Successfully cleaned up orphaned model versions archives.
                    Cleaning up deleted jobs from RQ user registries…
                    Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                    Removing unused worker types…
                    Successfully removed 0 unused worker types.
                    """
                ).strip()
            )

        with self.assertRaises(Image.DoesNotExist):
            image_no_element_old_1.refresh_from_db()
        with self.assertRaises(Image.DoesNotExist):
            image_no_element_old_2.refresh_from_db()
        image_with_element.refresh_from_db()
        image_no_element_new.refresh_from_db()

    @patch("arkindex.documents.management.commands.cleanup.s3")
    def test_cleanup_logs(self, cleanup_s3_mock, s3_mock, rq_mock):
        process = Process.objects.create(
            farm=Farm.objects.first(),
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.superuser,
        )
        task = process.tasks.create(run=0, depth=0, slug="task", ttl=0)

        good_s3_log = MagicMock()
        good_s3_log.key = f"{task.id}.log"
        orphan_s3_log = MagicMock()
        orphan_s3_log.key = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa.log"
        unsupported_s3_log = MagicMock()
        unsupported_s3_log.key = "cant_touch_this.txt.vbs"
        broken_s3_log = MagicMock()
        broken_s3_log.key = "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb.log"
        broken_s3_log.delete.side_effect = ClientError({"Error": {"Code": "500"}}, "delete_object")

        cleanup_s3_mock.Bucket.return_value.objects.all.side_effect = [
            # Bucket for Ponos artifacts
            [],
            # Bucket for corpus exports
            [],
            # Bucket for IIIF images
            [],
            # Bucket for Ponos logs
            [good_s3_log, orphan_s3_log, unsupported_s3_log, broken_s3_log],
            # Bucket for model version archives
            []
        ]

        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Removing log aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa.log…
                Unsupported log cant_touch_this.txt.vbs
                Removing log bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb.log…
                An error occurred (500) when calling the delete_object operation: Unknown
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        self.assertEqual(cleanup_s3_mock.Bucket.call_count, 5)
        self.assertListEqual(cleanup_s3_mock.Bucket.call_args_list, [
            call("ponos-artifacts"),
            call("export"),
            call("iiif"),
            call("ponos-logs"),
            call("training")
        ])
        self.assertEqual(cleanup_s3_mock.Bucket().objects.all.call_count, 5)
        self.assertEqual(good_s3_log.delete.call_count, 0)
        self.assertEqual(orphan_s3_log.delete.call_count, 1)
        self.assertEqual(unsupported_s3_log.delete.call_count, 0)
        self.assertEqual(broken_s3_log.delete.call_count, 1)

    @patch("arkindex.documents.management.commands.cleanup.s3")
    def test_cleanup_model_versions(self, cleanup_s3_mock, s3_mock, rq_mock):
        model = Model.objects.create(name="Nice Model")
        model_version = ModelVersion.objects.create(model_id=model.id, hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb", archive_hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", size=8)

        good_s3_version = MagicMock()
        good_s3_version.key = f"{model_version.id}.zst"
        orphan_s3_version = MagicMock()
        orphan_s3_version.key = "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb.zst"
        unsupported_s3_version = MagicMock()
        unsupported_s3_version.key = "cant_touch_this.txt.vbs"
        broken_s3_version = MagicMock()
        broken_s3_version.key = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa.zst"
        broken_s3_version.delete.side_effect = ClientError({"Error": {"Code": "500"}}, "delete_object")

        cleanup_s3_mock.Bucket.return_value.objects.all.side_effect = [
            # Bucket for Ponos artifacts
            [],
            # Bucket for corpus exports
            [],
            # Bucket for IIIF images
            [],
            # Bucket for Ponos logs
            [],
            # Bucket for model version archives
            [good_s3_version, orphan_s3_version, unsupported_s3_version, broken_s3_version]
        ]

        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Removing model version archive bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb.zst…
                Unsupported model version archive cant_touch_this.txt.vbs
                Removing model version archive aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa.zst…
                An error occurred (500) when calling the delete_object operation: Unknown
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        self.assertEqual(cleanup_s3_mock.Bucket.call_count, 5)
        self.assertListEqual(cleanup_s3_mock.Bucket.call_args_list, [
            call("ponos-artifacts"),
            call("export"),
            call("iiif"),
            call("ponos-logs"),
            call("training")
        ])
        self.assertEqual(cleanup_s3_mock.Bucket().objects.all.call_count, 5)
        self.assertEqual(good_s3_version.delete.call_count, 0)
        self.assertEqual(orphan_s3_version.delete.call_count, 1)
        self.assertEqual(unsupported_s3_version.delete.call_count, 0)
        self.assertEqual(broken_s3_version.delete.call_count, 1)

    @patch("arkindex.documents.management.commands.cleanup.s3")
    def test_cleanup_orphan_exports(self, cleanup_s3_mock, s3_mock, rq_mock):
        export = self.corpus.exports.create(user=self.superuser)

        good_export = MagicMock()
        good_export.key = str(export.id)
        orphan_export = MagicMock()
        orphan_export.key = "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"
        unsupported_export = MagicMock()
        unsupported_export.key = "cant_touch_this.txt.vbs"
        broken_export = MagicMock()
        broken_export.key = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
        broken_export.delete.side_effect = ClientError({"Error": {"Code": "500"}}, "delete_object")

        cleanup_s3_mock.Bucket.return_value.objects.all.side_effect = [
            # Bucket for Ponos artifacts
            [],
            # Bucket for corpus exports
            [good_export, orphan_export, unsupported_export, broken_export],
            # Bucket for IIIF images
            [],
            # Bucket for Ponos logs
            [],
            # Bucket for model version archives
            [],
        ]

        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Removing corpus export bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb…
                Unsupported corpus export cant_touch_this.txt.vbs
                Removing corpus export aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa…
                An error occurred (500) when calling the delete_object operation: Unknown
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        self.assertEqual(cleanup_s3_mock.Bucket.call_count, 5)
        self.assertListEqual(cleanup_s3_mock.Bucket.call_args_list, [
            call("ponos-artifacts"),
            call("export"),
            call("iiif"),
            call("ponos-logs"),
            call("training")
        ])
        self.assertEqual(cleanup_s3_mock.Bucket().objects.all.call_count, 5)
        self.assertEqual(good_export.delete.call_count, 0)
        self.assertEqual(orphan_export.delete.call_count, 1)
        self.assertEqual(unsupported_export.delete.call_count, 0)
        self.assertEqual(broken_export.delete.call_count, 1)

    @override_settings(REDIS_ZREM_CHUNK_SIZE=2)
    def test_cleanup_rq_user_registries(self, s3_mock, rq_mock):
        # List of user registries
        rq_mock().keys.return_value = [
            "rq:registry:user:1",
            "rq:registry:user:2",
            "rq:registry:user:3",
        ]
        # All keys in each user registry, in the order specified above
        rq_mock().zrange.side_effect = [
            [],
            ["job1", "job2", "job3", "job4", "job5"],
            ["job2", "job3"],
        ]
        # Whether or not each job exists in each non-empty user registry, in the order specified above
        rq_mock().pipeline().__enter__().execute.side_effect = [
            # Three jobs should be deleted: two ZREM calls will be made since we use chunks of 2
            [True, False, True, False, False],
            # Nothing to delete, no ZREM should be called
            [True, True],
        ]
        # How many jobs get deleted with each ZREM call
        rq_mock().zrem.side_effect = [2, 1]
        # Reset the call history, keeping the return values and side effects in place
        rq_mock.reset_mock()

        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 3 deleted jobs from 2 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        self.assertEqual(rq_mock.call_count, 1)
        self.assertEqual(rq_mock().keys.call_count, 1)
        self.assertEqual(rq_mock().keys.call_args, call(pattern="rq:registry:user:*"))
        self.assertListEqual(rq_mock().zrange.call_args_list, [
            call("rq:registry:user:1", 0, -1),
            call("rq:registry:user:2", 0, -1),
            call("rq:registry:user:3", 0, -1),
        ])
        self.assertEqual(rq_mock().pipeline.call_count, 2)
        self.assertListEqual(rq_mock().pipeline().__enter__().exists.call_args_list, [
            call(b"rq:job:job1"),
            call(b"rq:job:job2"),
            call(b"rq:job:job3"),
            call(b"rq:job:job4"),
            call(b"rq:job:job5"),
            call(b"rq:job:job2"),
            call(b"rq:job:job3"),
        ])
        self.assertEqual(rq_mock().pipeline().__enter__().execute.call_count, 2)
        self.assertEqual(rq_mock().zrem.call_count, 2)
        chunk_call, remaining_call = rq_mock().zrem.call_args_list
        # The chunk call is made using a set(), because we do not want to send duplicate keys,
        # so the arguments might be either job2,job4 or job4,job2
        self.assertIn(chunk_call, (
            call("rq:registry:user:2", "job2", "job4"),
            call("rq:registry:user:2", "job4", "job2"),
        ))
        self.assertEqual(remaining_call, call("rq:registry:user:2", "job5"))

    @override_settings(WORKER_CLEANUP_DELAY=10)
    def test_cleanup_archived_workers(self, s3_mock, rq_mock):
        # This worker can be cleaned up
        removable_worker = Worker.objects.get(slug="dla")
        self.assertFalse(removable_worker.versions.all().in_use())
        removable_worker.archived = datetime.now(timezone.utc) - timedelta(days=11)
        removable_worker.save()

        # Create a process where a worker version is used both with and without a configuration from this worker
        # This could lead to integrity errors with the default Django removal, as the configuration would be removed
        # from the WorkerRuns, so there would be 2 runs with the same version and no configuration when they should be unique
        process = self.corpus.processes.create(mode=ProcessMode.Workers, creator=self.superuser)
        version = removable_worker.versions.first()
        process.worker_runs.create(version=version, ttl=0)
        process.worker_runs.create(
            version=version,
            configuration=removable_worker.configurations.create(
                name="Some configuration",
                configuration={},
            ),
            ttl=0,
        )

        # This worker cannot be cleaned up because it is used in ML results
        used_worker = Worker.objects.get(slug="reco")
        self.assertTrue(used_worker.versions.all().in_use())
        used_worker.archived = datetime.now(timezone.utc) - timedelta(days=11)
        used_worker.save()

        non_archived = Worker.objects.create(
            name="Unarchived worker",
            type=used_worker.type,
            slug="unarchived",
        )
        recently_archived = Worker.objects.create(
            name="Recently archived worker",
            type=used_worker.type,
            slug="pending",
            archived=datetime.now(timezone.utc) - timedelta(days=9),
        )

        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing archived workers…
                Removing worker Document layout analyser ({removable_worker.id})
                Skipping 1 archived workers that have worker versions or configurations used in worker results.
                Successfully cleaned up 1 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 1 unused worker types.
                """
            ).strip()
        )

        with self.assertRaises(Worker.DoesNotExist):
            removable_worker.refresh_from_db()
        used_worker.refresh_from_db()
        non_archived.refresh_from_db()
        recently_archived.refresh_from_db()

    @override_settings(MODEL_CLEANUP_DELAY=10)
    def test_cleanup_archived_models(self, s3_mock, rq_mock):
        # This model can be cleaned up
        removable_model = Model.objects.create(name="Removable", archived=datetime.now(timezone.utc) - timedelta(days=11))

        # This model cannot be cleaned up because it is used in ML results
        used_model = Model.objects.create(name="Used", archived=datetime.now(timezone.utc) - timedelta(days=11))
        process = self.corpus.processes.create(mode=ProcessMode.Workers, creator=self.superuser)
        worker_version = WorkerVersion.objects.first()
        worker_run = process.worker_runs.create(
            version=worker_version,
            model_version=used_model.versions.create(),
            ttl=0,
        )
        self.corpus.elements.create(
            type=self.corpus.types.first(),
            name="Some element",
            worker_run=worker_run,
            worker_version=worker_version,
        )
        self.assertTrue(WorkerRun.objects.filter(model_version__model=used_model).in_use())

        non_archived = Model.objects.create(name="Unarchived")
        recently_archived = Model.objects.create(
            name="Recently archived",
            archived=datetime.now(timezone.utc) - timedelta(days=9),
        )

        self.assertEqual(
            self.cleanup(),
            dedent(
                f"""
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Removing model Removable ({removable_model.id})
                Skipping 1 archived models that have model versions used in worker results.
                Successfully cleaned up 1 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 0 unused worker types.
                """
            ).strip()
        )

        with self.assertRaises(Model.DoesNotExist):
            removable_model.refresh_from_db()
        used_model.refresh_from_db()
        non_archived.refresh_from_db()
        recently_archived.refresh_from_db()


    def test_cleanup_worker_types(self, s3_mock, rq_mock):
        for worker_type in WorkerType.objects.all():
            self.assertTrue(Worker.objects.filter(type=worker_type).exists())
        unused_type = WorkerType.objects.create(slug="unused", display_name="A worker type that no worker uses")
        self.assertFalse(Worker.objects.filter(type=unused_type).exists())
        self.assertEqual(WorkerType.objects.count(), 8)

        self.assertEqual(
            self.cleanup(),
            dedent(
                """
                Removing archived workers…
                Successfully cleaned up 0 archived workers.
                Removing orphaned Ponos artifacts…
                Successfully cleaned up orphaned Ponos artifacts.
                Removing 0 artifacts of expired processes from S3…
                Removing logs for 0 tasks of expired processes from S3…
                Removing 0 artifacts of expired processes…
                Removing 0 tasks of expired processes…
                Successfully cleaned up expired processes.
                Removing 0 old corpus exports from S3…
                Removing 0 old corpus exports…
                Successfully cleaned up old corpus exports.
                Removing orphaned corpus exports…
                Successfully cleaned up orphaned corpus exports.
                Deleting 0 DataFiles marked as trashed from S3 and the database…
                Successfully cleaned up DataFiles marked as trashed.
                Removing orphan images…
                Successfully cleaned up orphan images.
                Removing orphaned local images…
                Successfully cleaned up orphaned local images.
                Removing orphaned Ponos logs…
                Successfully cleaned up orphaned Ponos logs.
                Removing archived models…
                Successfully cleaned up 0 archived models.
                Removing orphaned model versions archives…
                Successfully cleaned up orphaned model versions archives.
                Cleaning up deleted jobs from RQ user registries…
                Successfully cleaned up 0 deleted jobs from 0 RQ user registries.
                Removing unused worker types…
                Successfully removed 1 unused worker types.
                """
            ).strip()
        )

        with self.assertRaises(WorkerType.DoesNotExist):
            unused_type.refresh_from_db()
        self.assertEqual(WorkerType.objects.count(), 7)
