from unittest.mock import call, patch
from uuid import uuid4

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import State
from arkindex.process.models import Process, ProcessMode
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestProcessPartialUpdate(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user)
        cls.page_type = cls.corpus.types.get(slug="page")
        cls.ml_class = cls.corpus.ml_classes.create(name="clafoutis")
        other_corpus = Corpus.objects.create(name="Corpus 2")
        cls.other_ml_class = other_corpus.ml_classes.create(name="beignet")

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.patch(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_running(self):
        """
        A running process cannot be updated... (see following test)
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.process.tasks.update(state=State.Running)

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:process-details", kwargs={"pk": self.process.id}),
                {"element_name_contains": "something"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"__all__": ["Cannot edit a process while it is running"]})

    def test_running_only_name(self):
        """
        Only name attribute of a running process can be patched
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.process.tasks.update(state=State.Running)
        self.assertEqual(self.process.name, None)

        with self.assertNumQueries(11):
            response = self.client.patch(
                reverse("api:process-details", kwargs={"pk": self.process.id}),
                {"name": "newName"},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.process.refresh_from_db()
        self.assertEqual(self.process.name, "newName")

    def test_file_import_element(self):
        """
        A file import's element can be updated even while it is running
        """
        self.client.force_login(self.user)
        process = self.corpus.processes.create(
            mode=ProcessMode.Files,
            creator=self.user,
        )
        process.run()
        process.tasks.update(state=State.Running)
        self.assertIsNone(self.process.element)

        # Take any element with an image: this will cause extra queries to retrieve the Image and ImageServer
        element = self.corpus.elements.exclude(image_id=None).first()

        with self.assertNumQueries(12):
            response = self.client.patch(
                reverse("api:process-details", kwargs={"pk": process.id}),
                {"element_id": str(element.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(process.element, element)

    @patch("arkindex.project.mixins.get_max_level", return_value=None)
    def test_no_permission(self, get_max_level_mock):
        """
        A user cannot update a process linked to a corpus they do not have admin access to
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.patch(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "Not found."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    def test_stop(self):
        """
        A pending or running process can be stopped
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()

        for state in [State.Pending, State.Running]:
            with self.subTest(state=state):
                self.process.tasks.update(state=state)
                self.process.refresh_from_db()
                self.assertEqual(self.process.state, state)

                with self.assertNumQueries(13):
                    response = self.client.patch(
                        reverse("api:process-details", kwargs={"pk": self.process.id}),
                        {"state": "stopping"},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.process.refresh_from_db()
                self.assertEqual(self.process.state, State.Stopping)
                self.assertIsNone(self.process.finished)

    def test_unscheduled_stop(self):
        """
        An unscheduled process can be stopped
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.assertEqual(self.process.state, State.Unscheduled)

        with self.assertNumQueries(14):
            response = self.client.patch(
                reverse("api:process-details", kwargs={"pk": self.process.id}),
                {"state": "stopping"},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.process.refresh_from_db()
        # Since no agents need to do anything to stop the tasks, all the tasks are immediately marked as Stopped
        self.assertEqual(self.process.state, State.Stopped)
        # Since everything is stopped, `finished` should be set
        self.assertIsNotNone(self.process.finished)

    def test_state_only_stopping(self):
        """
        A process' state can only be updated to `stopping`
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.assertEqual(self.process.state, State.Unscheduled)

        for state in set(State) - {State.Stopping}:
            with self.subTest(state=state):
                with self.assertNumQueries(6):
                    response = self.client.patch(
                        reverse("api:process-details", kwargs={"pk": self.process.id}),
                        {"state": state.value},
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    "state": ["Can only change the state to 'stopping'"],
                })

                self.process.refresh_from_db()
                self.assertEqual(self.process.state, State.Unscheduled)

    def test_state_non_final(self):
        """
        Only a process in an unscheduled, pending or running state can be updated to `stopping`
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()

        for state in set(State) - {State.Unscheduled, State.Pending, State.Running}:
            with self.subTest(state=state):
                self.process.tasks.update(state=state)
                self.process.refresh_from_db()
                self.assertEqual(self.process.state, state)

                with self.assertNumQueries(6):
                    response = self.client.patch(
                        reverse("api:process-details", kwargs={"pk": self.process.id}),
                        {"state": "stopping"},
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    "state": [f"Cannot stop a {state.value} process"],
                })

                self.process.refresh_from_db()
                self.assertEqual(self.process.state, state)

    def test_wrong_mode(self):
        """
        Fields other than 'name' cannot be updated in processes that are not of Workers mode
        """
        self.client.force_login(self.superuser)
        forbidden_modes = list(ProcessMode)
        forbidden_modes.remove(ProcessMode.Workers)
        # Local processes cannot be accessed by anyone because they have no corpus and no revision,
        # not even instance admins, so they cause a 404 instead of 400
        forbidden_modes.remove(ProcessMode.Local)

        for mode in forbidden_modes:
            with self.subTest(mode=mode):
                process = Process(
                    mode=mode,
                    creator=self.user,
                    corpus=self.corpus,
                )
                process.save()

                response = self.client.patch(
                    reverse("api:process-details", kwargs={"pk": process.id}),
                    {"element_name_contains": "something"},
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    response.json(),
                    {"__all__": ["Only processes of mode Workers can be updated"]}
                )

    def test_only_name(self):
        """
        Whatever the process mode, except Local, name attribute (alone) can be updated
        """
        self.client.force_login(self.superuser)
        modes = list(ProcessMode)
        # Local processes cannot be accessed by anyone because they have no corpus and no revision,
        # not even instance admins, so they cause a 404
        modes.remove(ProcessMode.Local)

        for mode in modes:
            with self.subTest(mode=mode):
                process = Process(
                    mode=mode,
                    creator=self.user,
                    corpus=self.corpus,
                )
                process.save()

                self.assertEqual(process.name, None)
                response = self.client.patch(
                    reverse("api:process-details", kwargs={"pk": process.id}),
                    {
                        "name": "newName"
                    },
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                process.refresh_from_db()
                self.assertEqual(process.name, "newName")

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Guest.value)
    def test_corpus_no_write_right(self, get_max_level_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.patch(reverse("api:process-details", kwargs={"pk": self.process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have a sufficient access level to this process."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    def test_partial_update(self):
        """
        Process fields related to elements filtering can be updated
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)
        self.assertIsNone(process.name)
        self.assertIsNone(process.name_contains)
        self.assertIsNone(process.element_type)
        self.assertIsNone(process.ml_class)
        self.assertFalse(process.load_children)

        with self.assertNumQueries(10):
            response = self.client.patch(
                reverse("api:process-details", kwargs={"pk": process.id}),
                {
                    "name": "newName",
                    "element_name_contains": "AAA",
                    "element_type": "page",
                    "ml_class_id": str(self.ml_class.id),
                    "load_children": True,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(process.name, "newName")
        self.assertEqual(process.name_contains, "AAA")
        self.assertEqual(process.element_type, self.page_type)
        self.assertEqual(process.ml_class, self.ml_class)
        self.assertTrue(process.load_children)

    def test_reset_fields(self):
        self.client.force_login(self.user)
        process = Process.objects.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.user,
            name="newName",
            name_contains="AAA",
            element_type=self.page_type,
            load_children=True
        )

        with self.assertNumQueries(8):
            response = self.client.patch(
                reverse("api:process-details", kwargs={"pk": process.id}),
                {
                    "name": None,
                    "element_name_contains": None,
                    "element_type": None,
                    "load_children": False
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertTupleEqual(
            (process.name, process.name_contains, process.element_type, process.load_children),
            (None, None, None, False)
        )

    def test_wrong_type(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:process-details", kwargs={"pk": self.process.id}),
                {"element_type": "unexisting_type"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            response.json(),
            {"element_type": ["Object with slug=unexisting_type does not exist."]}
        )

    def test_invalid_ml_class(self):
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)

        non_existent_id = uuid4()
        cases = [
            ("je says pas", "“je says pas” is not a valid UUID.", 5),
            (non_existent_id, f'Invalid pk "{non_existent_id}" - object does not exist.', 6),
            (self.other_ml_class.id, f'Invalid pk "{self.other_ml_class.id}" - object does not exist.', 6),
        ]
        for ml_class_id, message, queries in cases:
            with self.subTest(ml_class_id=ml_class_id), self.assertNumQueries(queries):
                response = self.client.patch(
                    reverse("api:process-details", kwargs={"pk": process.id}),
                    {
                        "ml_class_id": str(ml_class_id),
                    },
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), {
                    "ml_class_id": [message],
                })

    def test_fields(self):
        """
        Ensure determined fields can be updated on a non started workers process.
        Editable fields are `name`, `name_contains`, `element_type`, and `load_children`.
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)

        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse("api:process-details", kwargs={"pk": process.id}),
                {
                    "id": str(uuid4()),
                    "element": str(uuid4()),
                    "workflow": str(uuid4()),
                    "creator": str(uuid4()),
                    "files": [1, 2, 3],
                    "mode": "iiif",
                    "chunks": 42,
                    "folder_type": "register",
                    "name": "newName",
                    "element_name_contains": "AAA",
                    "element_type": "page",
                    "load_children": True,
                    "use_cache": True,
                    "activity_state": "pending",
                    "created": "2038-01-19T03:14:08Z",
                    "updated": "2038-01-19T03:14:08Z",
                    "started": "2038-01-19T03:14:08Z",
                    "finished": "2038-01-19T03:14:08Z",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(response.json(), {
            "name": "newName",
            "element_name_contains": "AAA",
            "element_type": "page",
            "folder_type": None,
            "ml_class_id": None,
            "load_children": True,
            "use_cache": False,
            "activity_state": "disabled",
            "id": str(process.id),
            "corpus": str(self.corpus.id),
            "chunks": 1,
            "element": None,
            "files": [],
            "mode": "workers",
            "state": "unscheduled",
            "template_id": None,
            "farm": None,
            "tasks": [],
            "created": process.created.isoformat().replace("+00:00", "Z"),
            "updated": process.updated.isoformat().replace("+00:00", "Z"),
            "started": None,
            "finished": None,
        })
