from unittest.mock import patch
from uuid import uuid4

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import State
from arkindex.process.models import Process, ProcessMode
from arkindex.project.tests import FixtureAPITestCase


class TestProcessUpdate(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user)
        cls.ml_class = cls.corpus.ml_classes.create(name="clafoutis")
        other_corpus = Corpus.objects.create(name="Corpus 2")
        cls.other_ml_class = other_corpus.ml_classes.create(name="beignet")

    def test_wrong_mode(self):
        """
        Processes that are not of Workers mode cannot be updated, except the 'name' field
        """
        self.client.force_login(self.superuser)
        forbidden_modes = list(ProcessMode)
        forbidden_modes.remove(ProcessMode.Workers)
        # Local processes cannot be accessed by anyone because they have no corpus and no revision,
        # not even instance admins, so they cause a 404 instead of 400
        forbidden_modes.remove(ProcessMode.Local)

        for mode in forbidden_modes:
            with self.subTest(mode=mode):
                process = Process.objects.create(
                    mode=mode,
                    creator=self.user,
                    corpus=self.corpus,
                )

                with self.assertNumQueries(6):
                    response = self.client.put(
                        reverse("api:process-details", kwargs={"pk": process.id}),
                        {
                            "name": "newName",
                            "template_id": None,
                            "element_name_contains": "AAA",
                            "element_type": "page",
                            "load_children": True
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(
                    response.json(),
                    {"__all__": ["Only processes of mode Workers can be updated"]}
                )

    def test_unscheduled_stop(self):
        """
        An unscheduled process can be stopped
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.assertEqual(self.process.state, State.Unscheduled)

        with self.assertNumQueries(15):
            response = self.client.put(
                reverse("api:process-details", kwargs={"pk": self.process.id}),
                {
                    "name": "newName",
                    "template_id": None,
                    "element_name_contains": "AAA",
                    "element_type": "page",
                    "load_children": True,
                    "state": "stopping",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.process.refresh_from_db()
        # Since no agents need to do anything to stop the tasks, all the tasks are immediately marked as Stopped
        self.assertEqual(self.process.state, State.Stopped)
        # Since everything is stopped, `finished` should be set
        self.assertIsNotNone(self.process.finished)

    def test_state_only_stopping(self):
        """
        A process' state can only be updated to `stopping`
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.assertEqual(self.process.state, State.Unscheduled)

        for state in set(State) - {State.Stopping}:
            with self.subTest(state=state):
                with self.assertNumQueries(7):
                    response = self.client.put(
                        reverse("api:process-details", kwargs={"pk": self.process.id}),
                        {
                            "name": "newName",
                            "template_id": None,
                            "element_name_contains": "AAA",
                            "element_type": "page",
                            "load_children": True,
                            "state": state.value,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    "state": ["Can only change the state to 'stopping'"],
                })

                self.process.refresh_from_db()
                self.assertEqual(self.process.state, State.Unscheduled)

    def test_state_non_final(self):
        """
        Only a process in an unscheduled, pending or running state can be updated to `stopping`
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()

        for state in set(State) - {State.Unscheduled, State.Pending, State.Running}:
            with self.subTest(state=state):
                self.process.tasks.update(state=state)
                self.process.refresh_from_db()
                self.assertEqual(self.process.state, state)

                with self.assertNumQueries(7):
                    response = self.client.put(
                        reverse("api:process-details", kwargs={"pk": self.process.id}),
                        {
                            "name": "newName",
                            "template_id": None,
                            "element_name_contains": "AAA",
                            "element_type": "page",
                            "load_children": True,
                            "state": "stopping",
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    "state": [f"Cannot stop a {state.value} process"],
                })

                self.process.refresh_from_db()
                self.assertEqual(self.process.state, state)

    def test_update(self):
        """
        Process fields related to elements filtering can be updated
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)
        self.assertIsNone(process.name)
        self.assertIsNone(process.name_contains)
        self.assertIsNone(process.element_type)
        self.assertIsNone(process.ml_class)
        self.assertFalse(process.load_children)

        with self.assertNumQueries(10):
            response = self.client.put(
                reverse("api:process-details", kwargs={"pk": process.id}),
                {
                    "name": "newName",
                    "element_name_contains": "AAA",
                    "element_type": "page",
                    "ml_class_id": str(self.ml_class.id),
                    "load_children": True,
                    "template_id": None,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(process.name, "newName")
        self.assertEqual(process.name_contains, "AAA")
        self.assertEqual(process.element_type.corpus_id, self.corpus.id)
        self.assertEqual(process.element_type.slug, "page")
        self.assertEqual(process.ml_class, self.ml_class)
        self.assertTrue(process.load_children)

    def test_invalid_ml_class(self):
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)

        non_existent_id = uuid4()
        cases = [
            ("je says pas", "“je says pas” is not a valid UUID.", 6),
            (non_existent_id, f'Invalid pk "{non_existent_id}" - object does not exist.', 7),
            (self.other_ml_class.id, f'Invalid pk "{self.other_ml_class.id}" - object does not exist.', 7),
        ]
        for ml_class_id, message, queries in cases:
            with self.subTest(ml_class_id=ml_class_id), self.assertNumQueries(queries):
                response = self.client.put(
                    reverse("api:process-details", kwargs={"pk": process.id}),
                    {
                        "name": "newName",
                        "element_name_contains": "AAA",
                        "element_type": "page",
                        "ml_class_id": str(ml_class_id),
                        "load_children": True,
                        "template_id": None,
                    },
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), {
                    "ml_class_id": [message],
                })

    def test_fields(self):
        """
        Ensure determined fields can be updated on a non started workers process.
        Editable fields are `name`, `name_contains`, `element_type`, and `load_children`.
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)

        with self.assertNumQueries(9):
            response = self.client.put(
                reverse("api:process-details", kwargs={"pk": process.id}),
                {
                    "id": str(uuid4()),
                    "element": str(uuid4()),
                    "workflow": str(uuid4()),
                    "creator": str(uuid4()),
                    "files": [1, 2, 3],
                    "mode": "iiif",
                    "chunks": 42,
                    "folder_type": "register",
                    "name": "newName",
                    "element_name_contains": "AAA",
                    "element_type": "page",
                    "load_children": True,
                    "use_cache": True,
                    "activity_state": "pending",
                    "template_id": None,
                    "created": "2038-01-19T03:14:08Z",
                    "updated": "2038-01-19T03:14:08Z",
                    "started": "2038-01-19T03:14:08Z",
                    "finished": "2038-01-19T03:14:08Z",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(response.json(), {
            "name": "newName",
            "element_name_contains": "AAA",
            "element_type": "page",
            "farm": None,
            "folder_type": None,
            "ml_class_id": None,
            "load_children": True,
            "use_cache": False,
            "chunks": 1,
            "activity_state": "disabled",
            "id": str(process.id),
            "corpus": str(self.corpus.id),
            "element": None,
            "files": [],
            "mode": "workers",
            "state": "unscheduled",
            "template_id": None,
            "tasks": [],
            "created": process.created.isoformat().replace("+00:00", "Z"),
            "updated": process.updated.isoformat().replace("+00:00", "Z"),
            "started": None,
            "finished": None,
        })
